/*
  omron-hid.c - driver OMRON USB/HID devices with NUT
  Copyright (C) BUFFALO INC.
 */

#include <pthread.h>
#include <ctype.h>
#include "main.h"
#include "libusb.h"
#include "usb-common.h"
#include "omronups.h"


#define SUB_DRIVER_VERSION	"0.01"
#define SUB_DRIVER_NAME		"OMRON Serial-over-USB UPS driver"

upsdrv_info_t upsdrv_info = {
	.name		= SUB_DRIVER_NAME,
	.version	= SUB_DRIVER_VERSION,
	.authors	= "BUFFALO INC.",
	.status		= DRV_EXPERIMENTAL,
	.subdrv_info	= { NULL }
};


/* USB ManufacurerId for OMRON */
#define OMRON_VID	0x0590


/* USB ProductId for OMRON UPSs */
#define BZ35T		0x0050
#define BZ50T		0x004B
#define BZ50LT		0x004F

#define BX35F		0x0058
#define BX50F		0x0057

#define BY50FW		0x0065
#define BY75SW		0x0066

#define BN150XR		0x0044
#define BN240XR		0x0043
#define BN100XR		0x0054

#define BU1002SW	0x006B
#define BU3002SW	0x006C

#define BU100RW		0x0075
#define BU200RW		0x0076
#define BU300RW		0x0077

#define BN50S		0x006E
#define BN75S		0x006F
#define BN100S		0x0070
#define BN150S		0x0071
#define BN220S		0x0072
#define BN300S		0x0073


static usb_device_id_t omron_usb_device_table[] = {
	{ .vendorID = OMRON_VID, .productID = BZ35T },
	{ .vendorID = OMRON_VID, .productID = BZ50T },
	{ .vendorID = OMRON_VID, .productID = BZ50LT },
	{ .vendorID = OMRON_VID, .productID = BX35F },
	{ .vendorID = OMRON_VID, .productID = BX50F },
	{ .vendorID = OMRON_VID, .productID = BY50FW },
	{ .vendorID = OMRON_VID, .productID = BY75SW },
	{ .vendorID = OMRON_VID, .productID = BN150XR },
	{ .vendorID = OMRON_VID, .productID = BN240XR },
	{ .vendorID = OMRON_VID, .productID = BN100XR },
	{ .vendorID = OMRON_VID, .productID = BU1002SW },
	{ .vendorID = OMRON_VID, .productID = BU3002SW },
	{ .vendorID = OMRON_VID, .productID = BU100RW },
	{ .vendorID = OMRON_VID, .productID = BU200RW },
	{ .vendorID = OMRON_VID, .productID = BU300RW },
	{ .vendorID = OMRON_VID, .productID = BN50S },
	{ .vendorID = OMRON_VID, .productID = BN75S },
	{ .vendorID = OMRON_VID, .productID = BN100S },
	{ .vendorID = OMRON_VID, .productID = BN150S },
	{ .vendorID = OMRON_VID, .productID = BN220S },
	{ .vendorID = OMRON_VID, .productID = BN300S },
	{ .vendorID = -1,	 .productID = -1 }/* terminator */
};

static int shutdown_delay = 1;	/* minute */
static usb_communication_subdriver_t *usb = &usb_subdriver;
static usb_dev_handle *udev = NULL;
static USBDevice_t usbdevice;
static unsigned int comm_failures = 0;


#define USB_ERR_LIMIT		10
#define USB_ERR_RATE		10


struct {
	pthread_t tid;
	pthread_mutex_t lock;
	int drained;
} drain_thread_info = {
	.lock = PTHREAD_MUTEX_INITIALIZER,
	.drained = 0,
};

static void *drain_thread(void *arg)
{
	struct omronups_stat upsstat;
	int i;

	/*
	 * Read rubbish data a few times;
	 * the UPS doesn't seem to respond properly
	 * the first few times after connecting
	 */
	for (i = 0; i < 5; i ++) {
		omronups_query_stat(udev, &upsstat);
		sleep(1);
	}

	pthread_mutex_lock(&drain_thread_info.lock);
	drain_thread_info.drained = 1;
	pthread_mutex_unlock(&drain_thread_info.lock);

	return NULL;
}


static int read_serial_number(struct usb_device *dev, struct usb_bus *bus, char *buf, int bufsize)
{
	int busnum = bus->location;
	int devnum = dev->devnum;
	int vid = dev->descriptor.idVendor;
	int pid = dev->descriptor.idProduct;

	char *path = NULL;
	int fd = -1;
	int nread = 0;
	char *chr = NULL;

	size_t len = 0;
	size_t llen = 0;

	memset(buf, 0, bufsize);

	asprintf(&path, "/tmp/ups-%d:%d:%04x:%04x", busnum, devnum, vid, pid);
	if (path == NULL) {
		upslogx(LOG_ERR, "cannot allocate memory");
		goto done;
	}

	fd = open(path, O_RDONLY);
	if (fd < 0) {
		upslogx(LOG_ERR, "cannot open '%s'", path);
		goto done;
	}

	while (1) {
		char *buf_ = buf + nread;
		int bufsize_ = bufsize - nread - sizeof('\0');
		int nread_ = read(fd, buf_, bufsize_);
		if (nread_ < 1) { break; }
		nread += nread_;
	}
	//right trim length
	len = strlen(buf);
	while (len && !isgraph(buf[len-1])) --len;
	//left trim length
	while (len && !isgraph(buf[llen])) ++llen;
	//trim
	if (llen) {
		memmove(buf, buf + llen, len -= llen);
	}
	buf[len] = '\0';

done:
	if (path) { free(path); path = NULL; }
	if (fd >= 0) { close(fd); }

	return strlen(buf);
}


static int device_possibly_match_func(USBDevice_t *device, void *privdata)
{
	if (device->VendorID == OMRON_VID) {
		return POSSIBLY_SUPPORTED;
	}
	return  NOT_SUPPORTED;
}


static int device_match_func(USBDevice_t *device, void *privdata)
{
	int status;

	if (getval("serial")) {
		if (! device->Serial) {
			goto unsupported;
		}

		if (strcmp(device->Serial, getval("serial")) != 0) {
			goto unsupported;
		}
	}

	status = is_usb_device_supported(omron_usb_device_table,
					 device->VendorID, device->ProductID);

	if (status == SUPPORTED) {
		upslogx(LOG_INFO, "Supported OMRON USB UPS(PID=0x%04x)\n",
			device->ProductID);
		return 1;
	}

	status = device_possibly_match_func(device, privdata);
	if (status == POSSIBLY_SUPPORTED) {
		upslogx(LOG_INFO, "Unknown OMRON USB UPS(PID=0x%04x), "
			"but possibly supported\n", device->ProductID);
		return 1;
	}

unsupported:
	return 0;
}


static USBDeviceMatcher_t device_matcher = {
	.match_function = device_match_func
};


static void usb_comm_fail(const char *fmt, ...)
{
	int ret;
	char why[SMALLBUF];
	va_list ap;

	/* this means we're probably here because select was interrupted */
	if (exit_flag != 0) {
		return;	/* ignored, since we're about to exit anyway */
	}

	comm_failures ++;

	if ((comm_failures == USB_ERR_LIMIT) ||
	    ((comm_failures % USB_ERR_RATE) == 0)) {
		upslogx(LOG_WARNING, "Warning: excessive comm failures, "
			"limiting error reporting");
	}

	/* once it's past the limit, only log once every USB_ERR_LIMIT calls */
	if ((comm_failures > USB_ERR_LIMIT) &&
	    ((comm_failures % USB_ERR_LIMIT) != 0)) {
		return;
	}

	/* generic message if the caller hasn't elaborated */
	if (!fmt) {
		upslogx(LOG_WARNING,
			"Communications with UPS lost - check cabling");
		return;
	}

	va_start(ap, fmt);
	ret = vsnprintf(why, sizeof(why), fmt, ap);
	va_end(ap);

	if ((ret < 1) || (ret >= (int) sizeof(why))) {
		upslogx(LOG_WARNING,
			"usb_comm_fail: vsnprintf needed more than %d bytes",
			sizeof(why));
	}

	upslogx(LOG_WARNING, "Communications with UPS lost: %s", why);
}

static void usb_comm_good()
{
	if (comm_failures == 0) {
		return;
	}

	upslogx(LOG_NOTICE, "Communications with UPS re-established");
	comm_failures = 0;
}

/*
 * Callback that is called by usb_device_open() that handles USB device
 * settings prior to accepting the devide. At the very least claim the
 * device here. Detaching the kernel driver will be handled by the
 * caller, don't do this here. Return < 0 on error, 0 or higher on
 * success.
 */
static int driver_callback(usb_dev_handle *handle, USBDevice_t *device)
{
	if (usb_set_configuration(handle, 1) < 0) {
		upsdebugx(5, "Can't set USB configuration");
		return -1;
	}

	if (usb_claim_interface(handle, 0) < 0) {
		upsdebugx(5, "Can't claim USB interface");
		return -1;
	}

	(void) usb_set_altinterface(handle, 0);

	if (usb_clear_halt(handle, 0x81) < 0) {
		upsdebugx(5, "Can't reset USB endpoint");
		return -1;
	}

	return 1;
}

static int usb_device_close(usb_dev_handle *handle)
{
	if (!handle) {
		return 0;
	}

	/* usb_release_interface() sometimes blocks and goes
	into uninterruptible sleep.  So don't do it. */
	/* usb_release_interface(handle, 0); */
	return usb_close(handle);
}

static int usb_device_open(usb_dev_handle **handlep, USBDevice_t *device,
			   USBDeviceMatcher_t *matcher,
			   int (*callback)(usb_dev_handle *handle,
					   USBDevice_t *device))
{
	struct usb_bus	*bus;

	/* libusb base init */
	usb_init();
	usb_find_busses();
	usb_find_devices();

#ifndef __linux__ /* SUN_LIBUSB (confirmed to work on Solaris and FreeBSD) */
	/* Causes a double free corruption in linux if device is detached! */
	usb_device_close(*handlep);
#endif

	for (bus = usb_busses; bus; bus = bus->next) {
		struct usb_device *dev;
		usb_dev_handle *handle;

		for (dev = bus->devices; dev; dev = dev->next) {
			int i, ret;
			USBDeviceMatcher_t *m;

			upsdebugx(4, "Checking USB device [%04x:%04x] (%s/%s)",
				  dev->descriptor.idVendor,
				  dev->descriptor.idProduct, bus->dirname,
				  dev->filename);
			
			/* supported vendors are now checked
			   by the supplied matcher */

			/* open the device */
			*handlep = handle = usb_open(dev);
			if (!handle) {
				upsdebugx(4, "Failed to open USB device, "
					  "skipping: %s", usb_strerror());
				continue;
			}

			/* collect the identifying information of this
			   device. Note that this is safe, because
			   there's no need to claim an interface for
			   this (and therefore we do not yet need to
			   detach any kernel drivers). */

			free(device->Vendor);
			free(device->Product);
			free(device->Serial); device->Serial = NULL;
			free(device->Bus);

			memset(device, 0, sizeof(*device));

			device->VendorID = dev->descriptor.idVendor;
			device->ProductID = dev->descriptor.idProduct;
			device->Bus = strdup(bus->dirname);

			if (dev->descriptor.iManufacturer) {
				char	buf[SMALLBUF];
				ret = usb_get_string_simple(handle,
						dev->descriptor.iManufacturer,
							    buf, sizeof(buf));
				if (ret > 0) {
					device->Vendor = strdup(buf);
				}
			}

			if (dev->descriptor.iProduct) {
				char	buf[SMALLBUF];
				ret = usb_get_string_simple(handle,
						dev->descriptor.iProduct,
							    buf, sizeof(buf));
				if (ret > 0) {
					device->Product = strdup(buf);
				}
			}

			if (dev->descriptor.iSerialNumber) {
				char	buf[SMALLBUF];
				ret = usb_get_string_simple(handle,
						dev->descriptor.iSerialNumber,
							    buf, sizeof(buf));
				if (ret > 0) {
					device->Serial = strdup(buf);
				}
			}
			if (!device->Serial) {
				char buf[SMALLBUF];
				ret = read_serial_number(dev, bus, buf, sizeof(buf));
				if (ret > 0) {
					device->Serial = strdup(buf);
				}
			}

			upsdebugx(4, "- VendorID     : %04x",
				  device->VendorID);
			upsdebugx(4, "- ProductID    : %04x",
				  device->ProductID);
			upsdebugx(4, "- Manufacturer : %s",
				  device->Vendor ? device->Vendor : "unknown");
			upsdebugx(4, "- Product      : %s",
				  device->Product ? device->Product : "unknown");
			upsdebugx(4, "- Serial Number: %s",
				  device->Serial ? device->Serial : "unknown");
			upsdebugx(4, "- Bus          : %s",
				  device->Bus ? device->Bus : "unknown");

			for (m = matcher; m; m = m->next) {
				
				switch (m->match_function(device, m->privdata))
				{
				case 0:
					upsdebugx(4, "Device does not match "
						  "- skipping");
					goto next_device;
				case -1:
					fatal_with_errno(EXIT_FAILURE,
							 "matcher");
					goto next_device;
				case -2:
					upsdebugx(4, "matcher: unspecified error");
					goto next_device;
				}
			}

			for (i = 0; i < 3; i ++) {

				ret = callback(handle, device);
				if (ret >= 0) {
					upsdebugx(4, "USB device [%04x:%04x] opened",
						  device->VendorID,
						  device->ProductID);
					return ret;
				}
#if HAVE_USB_DETACH_KERNEL_DRIVER_NP
				/* this method requires at least libusb 0.1.8:
				 * it force device claiming by unbinding
				 * attached driver... From libhid */
				if (usb_detach_kernel_driver_np(handle, 0) < 0) {
					upsdebugx(4, "failed to detach kernel "
						  "driver from USB device: %s",
						  usb_strerror());
				} else {
					upsdebugx(4, "detached kernel driver "
						  "from USB device...");
				}
#endif
			}

			fatalx(EXIT_FAILURE, "USB device [%04x:%04x] matches, "
			       "but driver callback failed: %s",
			       device->VendorID, device->ProductID,
			       usb_strerror());

		next_device:
			usb_close(handle);
		}
	}

	*handlep = NULL;
	upsdebugx(4, "No matching USB device found");

	return -1;
}

/*
 * Initialise the UPS
 */
void upsdrv_initups(void)
{
	int	i;

	for (i = 0; usb_device_open(&udev, &usbdevice, &device_matcher,
				    &driver_callback) < 0; i ++) {

		if ((i < 32) && (sleep(5) == 0)) {
			usb_comm_fail("Can't open USB device, retrying ...");
			continue;
		}

		fatalx(EXIT_FAILURE,
		       "Unable to find OMRON USB solution\n\n"

		       "Things to try:\n"
		       " - Connect UPS device to USB bus\n"
		       " - Run this driver as another user "
		       "(upsdrvctl -u or 'user=...' in ups.conf).\n"
		       "   See upsdrvctl(8) and ups.conf(5).\n\n"

		       "Fatal error: unusable configuration");
	}

	if (getval("sd_delay")) {
		shutdown_delay = atoi(getval("sd_delay"));
		upslogx(LOG_INFO, "shutdown delay = %d(minute)\n",
			shutdown_delay);
	}
}

void upsdrv_cleanup(void)
{
	usb_device_close(udev);

	free(usbdevice.Vendor);
	free(usbdevice.Product);
	free(usbdevice.Serial);
	free(usbdevice.Bus);

	pthread_mutex_destroy(&drain_thread_info.lock);
	pthread_join(drain_thread_info.tid, NULL);
}

void upsdrv_initinfo(void)
{
	dstate_setinfo("ups.mfr", "%s", "OMRON");
	dstate_setinfo("ups.model", "%s",
		       usbdevice.Product ? usbdevice.Product : "unknown");
	dstate_setinfo("ups.serial", "%s",
		       usbdevice.Serial ? usbdevice.Serial : "unknown");

	dstate_setinfo("ups.vendorid", "%04x", usbdevice.VendorID);
	dstate_setinfo("ups.productid", "%04x", usbdevice.ProductID);
}

void upsdrv_updateinfo(void)
{
	int	ret;
	struct omronups_stat upsstat;
	static int ncall = 0;

	if (!udev) {
		ret = usb_device_open(&udev, &usbdevice,
				      &device_matcher, &driver_callback);

		if (ret < 0) {
			return;
		}
	}

	if (ncall == 0) {
		/* upsdrv_updateinfo() has been called by main() before background() */
		++ncall;
	} else if (ncall == 1) {
		/* upsdrv_updateinfo() has been called by main() after background() */
		if (pthread_create(&drain_thread_info.tid, NULL, drain_thread, NULL) != 0) {
			fatalx(EXIT_FAILURE,
			       "Fatal error: unable to create rubbish data drain thread");
		}
		++ncall;
	}

	/*
	 * when rubbish data has not been drained, set dummy status to
	 * avoid E10 notification caused by upsmon
	 */
	pthread_mutex_lock(&drain_thread_info.lock);
	if (drain_thread_info.drained == 0) {
		pthread_mutex_unlock(&drain_thread_info.lock);
		upslogx(LOG_INFO, "rubbish data has not been drained");
		upsstat.ac_stat = ONLINE;
		upsstat.lo_batt = 0;
		goto done;
	}
	pthread_mutex_unlock(&drain_thread_info.lock);

	ret = omronups_query_stat(udev, &upsstat);

	if (ret < 4) {
		usb_comm_fail("Query to UPS failed");
		dstate_datastale();

		usb_device_close(udev);
		udev = NULL;

		return;
	}

done:
	usb_comm_good();
	dstate_dataok();

	status_init();

	if (upsstat.ac_stat == ONLINE) {
	    status_set("OL");
	}
	else {
	    status_set("OB");
	}

	if (upsstat.lo_batt) {
	    status_set("LB");
	}

	status_commit();

	upsdebugx(5, "ups.status: %s %s\n",
		  (upsstat.ac_stat == ONLINE) ? "OL" : "OB",
		  upsstat.lo_batt ? "" : "BL");
}

void upsdrv_shutdown(void)
{
	omronups_initiate_shutdown(udev, shutdown_delay);
}

void upsdrv_help(void)
{
}

void upsdrv_makevartable(void)
{
	addvar(VAR_VALUE, "sd_delay", "UPS shutdown after the time(minute)");
	addvar(VAR_VALUE, "serial", "UPS serial number");
}
